from nilmdb.utils.printf import *
import threading
import warnings

def verify_proxy(obj, exception = False,
                 check_thread = True, check_concurrent = True):
    """Return a VerifyObjectProxy that proxies all method calls
    to the given object, as well as attribute retrievals.

    When calling methods, the following checks are performed.
    If exception is True, an exception is raised.  Otherwise, a warning
    is printed.

    check_thread = True     # Warn/fail if two different threads call methods.
    check_concurrent = True # Warn/fail if two functions are concurrently
                            # run through this proxy

    Note that the __init__ call is not included in the checks, since
    this wrapper is passed an already instantiated object.
    """

    class VerifyCallProxy(object):
        def __init__(self, func, parent):
            self.func = func
            self.parent = parent

        def __call__(self, *args, **kwargs):
            p = self.parent
            this = threading.current_thread()
            callee = self.func.__name__
            classname = p.__class__.__name__

            if p.thread is None:
                p.thread = this
                p.thread_callee = callee

            if check_thread and p.thread != this:
                err = sprintf("unsafe threading: %s called %s.%s,"
                              " but %s called %s.%s",
                              p.thread.name, classname, p.thread_callee,
                              this.name, classname, callee)
                if exception:
                    raise AssertionError(err)
                else: # pragma: no cover
                    warnings.warn(err)

            need_concur_unlock = False
            if check_concurrent:
                if p.concur_lock.acquire(False) == False:
                    err = sprintf("unsafe concurrency: %s called %s.%s "
                                  "while %s is still in %s.%s",
                                  this.name, classname, callee,
                                  p.concur_tname, classname, p.concur_callee)
                    if exception:
                        raise AssertionError(err)
                    else: # pragma: no cover
                        warnings.warn(err)
                else:
                    p.concur_tname = this.name
                    p.concur_callee = callee
                    need_concur_unlock = True

            try:
                ret = self.func(*args, **kwargs)
            finally:
                if need_concur_unlock:
                    p.concur_lock.release()
            return ret

    class VerifyObjectProxy(object):
        def __init__(self, obj):
            self.obj = obj
            self.thread = None
            self.thread_callee = None
            self.concur_lock = threading.Lock()
            self.concur_tname = None
            self.concur_callee = None

        def __getattr__(self, key):
            attr = getattr(self.obj, key)
            if not callable(attr):
                return VerifyCallProxy(getattr, self)(self.obj, key)
            return VerifyCallProxy(attr, self)

    return VerifyObjectProxy(obj)
